package dev.flyfish.framework.r2dbc.repository.impl;

import dev.flyfish.framework.domain.base.Domain;
import dev.flyfish.framework.domain.base.Qo;
import dev.flyfish.framework.repository.DefaultReactiveRepository;
import lombok.Getter;
import org.apache.commons.lang3.StringUtils;
import org.reactivestreams.Publisher;
import org.springframework.dao.IncorrectResultSizeDataAccessException;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.data.r2dbc.convert.R2dbcConverter;
import org.springframework.data.r2dbc.core.R2dbcEntityOperations;
import org.springframework.data.r2dbc.repository.support.SimpleR2dbcRepository;
import org.springframework.data.relational.core.query.Criteria;
import org.springframework.data.relational.core.query.Query;
import org.springframework.data.relational.repository.query.RelationalEntityInformation;
import org.springframework.data.repository.core.EntityInformation;
import org.springframework.lang.NonNull;
import org.springframework.util.Assert;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.util.List;
import java.util.Optional;

/**
 * 默认的异步仓库实现
 *
 * @param <T> 泛型
 */
public class QueryingR2dbcReactiveRepository<T extends Domain> extends SimpleR2dbcRepository<T, String>
        implements DefaultReactiveRepository<T> {

    @Getter
    private final RelationalEntityInformation<T, String> entity;

    private final R2dbcEntityOperations entityOperations;

    /**
     * 构造新的默认仓库
     *
     * @param entity           实体信息
     * @param entityOperations 数据库操作模板
     * @param converter        转换器
     */
    public QueryingR2dbcReactiveRepository(RelationalEntityInformation<T, String> entity, R2dbcEntityOperations entityOperations, R2dbcConverter converter) {
        super(entity, entityOperations, converter);
        this.entity = entity;
        this.entityOperations = entityOperations;
    }

    /**
     * 通过名称查找一个
     *
     * @param name 名称
     * @return 结果
     */
    @Override
    public Mono<T> findByName(String name) {
        if (StringUtils.isNotBlank(name)) {
            return entityOperations
                    .selectOne(Query.query(Criteria.where("name").is(name)), entity.getJavaType());
        }
        return Mono.empty();
    }

    /**
     * Returns a single entity matching the given {@link Qo} or {@link Optional#empty()} if none was found.
     *
     * @param query must not be {@literal null}.
     * @return a single entity matching the given {@link Qo} or {@link Optional#empty()} if none was found.
     * @throws IncorrectResultSizeDataAccessException if the Qo yields more than one
     *                                                result.
     */
    @Override
    public Mono<T> findOne(Qo<T> query) {
        return getQuery(query)
                .flatMap(querying -> entityOperations.selectOne(querying, entity.getJavaType()))
                .contextWrite(ctx -> ctx.put(Qo.KEY, query));
    }

    /**
     * Returns all entities matching the given {@link Qo}. In case no match could be found an empty
     * {@link Iterable} is returned.
     *
     * @param query must not be {@literal null}.
     * @return all entities matching the given {@link Qo}.
     */
    @Override
    public Flux<T> findAll(Qo<T> query) {
        return getQuery(query)
                .flatMapMany(querying -> entityOperations.select(querying, entity.getJavaType()))
                .contextWrite(ctx -> ctx.put(Qo.KEY, query));
    }

    /**
     * Returns all entities matching the given {@link Qo} applying the given {@link Sort}. In case no match could
     * be found an empty {@link Iterable} is returned.
     *
     * @param query must not be {@literal null}.
     * @param sort  the {@link Sort} specification to sort the results by, may be {@link Sort#empty()}, must not be
     *              {@literal null}.
     * @return all entities matching the given {@link Qo}.
     * @since 1.10
     */
    @Override
    public Flux<T> findAll(Qo<T> query, Sort sort) {
        return getQuery(query)
                .flatMapMany(querying -> entityOperations.select(querying.sort(sort), entity.getJavaType()))
                .contextWrite(ctx -> ctx.put(Qo.KEY, query));
    }

    /**
     * Returns a {@link Page} of entities matching the given {@link Qo}. In case no match could be found, an empty
     * {@link Page} is returned.
     *
     * @param query    must not be {@literal null}.
     * @param pageable may be {@link Pageable#unpaged()}, must not be {@literal null}.
     * @return a {@link Page} of entities matching the given {@link Qo}.
     */
    @Override
    public Mono<Page<T>> findAll(Qo<T> query, Pageable pageable) {
        return getQuery(query)
                .flatMap(querying -> entityOperations.select(querying.with(pageable),
                                entity.getJavaType())
                        .collectList()
                        .flatMap(list -> ReactivePageableExecutionUtils.getPage(list, pageable, this.count(query))))
                .defaultIfEmpty(Page.empty())
                .contextWrite(ctx -> ctx.put(Qo.KEY, query));
    }

    /**
     * Returns the number of instances matching the given {@link Qo}.
     *
     * @param query the {@link Qo} to count instances for, must not be {@literal null}.
     * @return the number of instances matching the {@link Qo}.
     */
    @Override
    public Mono<Long> count(Qo<T> query) {
        return getQuery(query)
                .flatMap(querying -> this.entityOperations.count(querying, entity.getJavaType()))
                .defaultIfEmpty(0L);
    }

    /**
     * 通过特定键的集合查询
     *
     * @param key    键
     * @param values 集合
     * @return 结果
     */
    @Override
    public Flux<T> findAllByValues(String key, List<?> values) {
        Criteria criteria = Criteria.where(key).in(values);
        Query query = Query.query(criteria);
        return entityOperations.select(query, entity.getJavaType());
    }

    /**
     * Checks whether the data store contains elements that match the given {@link Qo}.
     *
     * @param query the {@link Qo} to use for the existence check, must not be {@literal null}.
     * @return {@literal true} if the data store contains elements that match the given {@link Qo}.
     */
    @Override
    public Mono<Boolean> exists(Qo<T> query) {
        return getQuery(query)
                .flatMap(querying -> entityOperations.exists(querying, entity.getJavaType()))
                .defaultIfEmpty(false);
    }

    /**
     * 删除全部
     *
     * @param query 查询实体
     * @return 结果
     */
    @Override
    public Mono<Void> deleteAll(Qo<T> query) {
        return getQuery(query)
                .flatMap(querying -> entityOperations.delete(querying, entity.getJavaType()))
                .then(Mono.empty());
    }

    /**
     * 插入给定的实体. 假设该实体是新的，以便应用插入优化。
     * 使用返回的实例进行后续操作，因为保存操作可能会完全改变实体实例。
     *
     * @param entity 要插入的实体，不能为空
     * @return 保存后的实体
     */
    @Override
    @NonNull
    public <S extends T> Mono<S> insert(@NonNull S entity) {
        Assert.notNull(entity, "Entity must not be null!");
        return entityOperations.insert(entity);
    }

    /**
     * 插入给定的实体列表. 假设这些实体是新的，以便应用插入优化。
     * 使用返回的实例进行后续操作，因为保存操作可能会完全改变实体实例。
     *
     * @param entities 实体列表，不能为空
     * @return 保存后的实体列表
     */
    @Override
    public <S extends T> Flux<S> insert(Iterable<S> entities) {
        Assert.notNull(entities, "Objects to save must not be null!");
        return Flux.fromIterable(entities).concatMap(this::insert);
    }

    /**
     * 插入给定的实体发布者发布的实体. 假设这些实体是新的，以便应用插入优化。
     * 使用返回的实例进行后续操作，因为保存操作可能会完全改变实体实例。
     *
     * @param entities 发布的实体列表，不能为空
     * @return 保存后的实体列表
     */
    @Override
    public <S extends T> Flux<S> insert(Publisher<S> entities) {
        Assert.notNull(entities, "The given Publisher of entities must not be null!");
        return Flux.from(entities).flatMap(this::insert);
    }

    @Override
    public EntityInformation<T, String> getEntityInformation() {
        return entity;
    }

    private Mono<Query> getQuery(Qo<T> qo) {
        return Mono.justOrEmpty(qo.getQuery(entity));
    }
}
