package dev.flyfish.framework.beans.repository;

import org.apache.commons.lang3.ClassUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.FactoryBean;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.beans.factory.support.AbstractBeanDefinition;
import org.springframework.beans.factory.support.BeanDefinitionBuilder;
import org.springframework.beans.factory.support.BeanDefinitionRegistry;
import org.springframework.beans.factory.support.BeanNameGenerator;
import org.springframework.core.env.Environment;
import org.springframework.core.io.ResourceLoader;
import org.springframework.core.type.AnnotationMetadata;
import org.springframework.data.repository.config.*;

import java.lang.annotation.Annotation;
import java.util.Arrays;
import java.util.List;

/**
 * 简单的仓库注册器
 *
 * @author wangyu
 * 封装大部分逻辑
 */
class SimpleRepositoryRegistrar implements RepositoryRegistrar {

    private final RepositoryConfigurationExtension extension;
    private RepositoryConfigurationSource configurationSource;
    private CustomRepositoryBuilder builder;
    // 支持的注解类型
    private final Class<? extends Annotation> annotationType;
    // 支持的仓库类型
    private final List<Class<?>> supportedTypes;

    SimpleRepositoryRegistrar(Class<? extends Annotation> annotationType, RepositoryConfigurationExtension extension,
                              Class<?>... supportedTypes) {
        this.annotationType = annotationType;
        this.extension = extension;
        this.supportedTypes = Arrays.asList(supportedTypes);
    }

    /**
     * 初始化注册器，不初始化无法使用
     *
     * @param metadata       元数据
     * @param resourceLoader 资源加载器
     * @param environment    环境变量
     * @param registry       bean注册器
     * @param generator      生成器
     * @return 结果
     */
    @Override
    public RepositoryRegistrar init(
            AnnotationMetadata metadata, ResourceLoader resourceLoader, Environment environment,
            BeanDefinitionRegistry registry, BeanNameGenerator generator) {
        // repo生成的构建器
        this.configurationSource = new AnnotationRepositoryConfigurationSource(metadata,
                annotationType, resourceLoader, environment, registry, generator);
        // 构建器
        this.builder = new CustomRepositoryBuilder(registry, extension,
                configurationSource, resourceLoader, environment);
        return this;
    }

    /**
     * @param clazz 编译后的class
     */
    @Override
    public void register(Class<?> clazz) {
        if (supportedTypes.stream().anyMatch(type -> ClassUtils.isAssignable(clazz, type))) {
            register(clazz, configurationSource, builder);
        }
    }

    /**
     * 注册仓库内部方法，区分异步和同步
     *
     * @param clazz               类型
     * @param configurationSource 配置源
     * @param builder             构建器
     */
    private void register(Class<?> clazz, RepositoryConfigurationSource configurationSource,
                          CustomRepositoryBuilder builder) {
        //生成BeanDefinition并注册到容器中
        BeanDefinition origin = BeanDefinitionBuilder.genericBeanDefinition(clazz).getRawBeanDefinition();
        //设置当前bean定义对象是单例的
        origin.setScope("singleton");
        RepositoryConfiguration<? extends RepositoryConfigurationSource> configuration =
                new DefaultRepositoryConfiguration<>(configurationSource, origin, extension);
        BeanDefinitionBuilder definitionBuilder = builder.build(configuration);
        extension.postProcess(definitionBuilder, configurationSource);
        extension.postProcess(definitionBuilder, (AnnotationRepositoryConfigurationSource) configurationSource);

        AbstractBeanDefinition beanDefinition = definitionBuilder.getBeanDefinition();
        beanDefinition.setPrimary(configuration.isPrimary());
        String beanName = StringUtils.uncapitalize(clazz.getSimpleName());
        beanDefinition.setAttribute(FactoryBean.OBJECT_TYPE_ATTRIBUTE, configuration.getRepositoryInterface());
        builder.getRegistry().registerBeanDefinition(beanName, beanDefinition);
    }
}
